﻿using IniFile;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Text;

namespace KEHmusic
{
	static class Program
	{
		static byte[] GetTextFileBytes(string path)
		{
			// reads a text file and outputs a string that line endings normalized to \n
			string[] lines = File.ReadAllLines(path, Encoding.Default);
			using(MemoryStream ms = new MemoryStream())
			{
				for (int i = 0; i < lines.Length; i ++)
				{
					byte[] line = Encoding.Default.GetBytes(lines[i] + '\n');
					ms.Write(line, 0, line.Length);
				}
				return ms.ToArray();
			}
		}

		static string GetFileHash(System.Security.Cryptography.HashAlgorithm hashAlgo, string path, bool isText)
		{
			byte[] fileData = isText ? GetTextFileBytes(path) : File.ReadAllBytes(path);
			byte[] hashbytes = hashAlgo.ComputeHash(fileData);
			string hashstr = string.Empty;
			foreach (byte b in hashbytes)
				hashstr += b.ToString("x2");
			return hashstr;
		}

		static void Main(string[] args)
		{
			var dacs = IniSerializer.Deserialize<Dictionary<string, DACInfo>>("sound/DAC/DAC.ini");
			var samples = new List<KeyValuePair<string, int>>();
			foreach (var item in dacs.Select(a => a.Value.File).Distinct())
				samples.Add(new KeyValuePair<string, int>(item, (int)new FileInfo(Path.Combine("sound/DAC", item)).Length));
			List<int> dacbanksizes = new List<int>() { 0 };
			List<List<string>> dacbanks = new List<List<string>> { new List<string>() };
			foreach (var item in samples.OrderByDescending(a => a.Value))
			{
				bool found = false;
				for (int bank = 0; bank < dacbanksizes.Count; bank++)
					if (dacbanksizes[bank] + item.Value <= 0x8000)
					{
						dacbanks[bank].Add(item.Key);
						dacbanksizes[bank] += item.Value;
						found = true;
						break;
					}
				if (!found)
				{
					dacbanks.Add(new List<string>() { item.Key });
					dacbanksizes.Add(item.Value);
				}
			}
			var samplenums = new Dictionary<string, string>();
			using (StreamWriter sw = new StreamWriter("dacbanks.asm", false, Encoding.ASCII))
			{
				int i = 1;
				for (int b = 0; b < dacbanks.Count; b++)
				{
					sw.WriteLine("SndDAC{0}_Start:\tstartBank", b + 1);
					sw.WriteLine();
					foreach (string item in dacbanks[b])
					{
						samplenums.Add(item, "SndDAC_Sample" + i);
						sw.WriteLine("SndDAC_Sample{0}:\tBINCLUDE \"sound/DAC/{1}\"", i, item);
						sw.WriteLine("SndDAC_Sample{0}_End:", i++);
						sw.WriteLine();
					}
					sw.WriteLine("\tfinishBank");
					sw.WriteLine();
				}
			}
			List<string> dacids = new List<string>(dacs.Count);
			using (StreamWriter sw = new StreamWriter("dacinfo.asm", false, Encoding.ASCII))
			{
				sw.WriteLine("zDACMasterPlaylist:");
				sw.WriteLine();
				foreach (var item in dacs)
				{
					string delay;
					if (item.Value.Delay >= 0xA0)
						delay = item.Value.Delay.ToString("X3") + "h";
					else if (item.Value.Delay >= 0xA)
						delay = item.Value.Delay.ToString("X2") + "h";
					else
						delay = item.Value.Delay.ToString();
					sw.WriteLine("\tDACSample\t{0},{1} ; {2}", samplenums[item.Value.File], delay, item.Key);
					dacids.Add(item.Key);
				}
			}
			using (StreamWriter sw = new StreamWriter("dacids.asm", false, Encoding.ASCII))
			{
				string last = "$81";
				for (int i = 0; i < dacids.Count; i++)
				{
					if (i % 7 == 0)
						sw.Write("\tenum {0}={1}", dacids[i], last);
					else
						sw.Write(",{0}", dacids[i]);
					if (i % 7 == 6)
					{
						sw.WriteLine();
						last = dacids[i] + "+1";
					}
				}
			}
			var songs = IniSerializer.Deserialize<Dictionary<string, MusicInfo>>("sound/music/music.ini");
			bool writeini = false;
			System.Security.Cryptography.MD5 md5hasher = System.Security.Cryptography.MD5.Create();
			foreach (var item in songs)
			{
				switch (Path.GetExtension(item.Value.File).ToLowerInvariant())
				{
					case ".asm":
					case ".68k":
						string md5 = GetFileHash(md5hasher, Path.Combine("sound/music", item.Value.File), true);
						if (item.Value.MD5 != md5 || item.Value.Size <= 0 || (item.Value.OutputFile != null && !File.Exists(Path.Combine("sound/music", item.Value.OutputFile))))
						{
							using (StreamWriter sw = new StreamWriter("tmp.asm", false, Encoding.ASCII))
							{
								sw.WriteLine("\tCPU 68000");
								sw.WriteLine("\tpadding off");
								sw.WriteLine("kehmusic = 1");
								sw.WriteLine("allOptimizations = 1");
								sw.WriteLine("\tinclude \"s2.macros.asm\"");
								sw.WriteLine("\tinclude \"sound/_smps2asm_inc.asm\"");
								if (item.Value.Compressible)
									sw.WriteLine("\tphase $1380");
								sw.WriteLine("\tinclude \"sound/music/{0}\"", item.Value.File);
							}
							ProcessStartInfo si = new ProcessStartInfo("win32/asw", "-E -xx -A -r 2 -q -U tmp.asm");
							si.CreateNoWindow = true;
							si.EnvironmentVariables.Add("AS_MSGPATH", "win32/msg");
							si.EnvironmentVariables.Add("USEANSI", "n");
							si.UseShellExecute = false;
							using (Process proc = Process.Start(si))
								proc.WaitForExit();
							File.Delete("tmp.asm");
							if (File.Exists("tmp.log"))
							{
								Console.Write(File.ReadAllText("tmp.log"));
								File.Delete("tmp.log");
							}
							if (!File.Exists("tmp.p")) continue;
							si = new ProcessStartInfo("win32/s2p2bin_plus", "tmp.p tmp.bin");
							si.CreateNoWindow = true;
							si.UseShellExecute = false;
							using (Process proc = Process.Start(si))
								proc.WaitForExit();
							File.Delete("tmp.p");
							if (!File.Exists("tmp.bin")) continue;
							item.Value.Size = (short)new FileInfo("tmp.bin").Length;
							item.Value.MD5 = md5;
							if (item.Value.Compressible && item.Value.Size <= 0x800)
							{
								item.Value.OutputFile = Path.GetFileNameWithoutExtension(item.Value.File) + "_cmp.bin";
								SonicRetro.KensSharp.Saxman.Compress("tmp.bin", Path.Combine("sound/music", item.Value.OutputFile), true);
								item.Value.Size = (short)new FileInfo(Path.Combine("sound/music", item.Value.OutputFile)).Length;
							}
							File.Delete("tmp.bin");
							writeini = true;
						}
						break;
					case ".bin":
						short size = (short)new FileInfo(Path.Combine("sound/music", item.Value.File)).Length;
						//string md5 = GetFileHash(md5hasher, Path.Combine("sound/music", item.Value.File), false);
						if (item.Value.Size != size)
						{
							item.Value.Size = size;
							writeini = true;
						}
						break;
				}
			}
			if (writeini)
				IniSerializer.Serialize(songs, "sound/music/music.ini");
			int bankmax = 0x8000 - (songs.Count * 3);
			List<int> banksizes = new List<int>() { 0 };
			List<List<KeyValuePair<string, MusicInfo>>> banks = new List<List<KeyValuePair<string, MusicInfo>>>() { new List<KeyValuePair<string, MusicInfo>>() };
			foreach (var item in songs.OrderByDescending(a => a.Value.Size))
			{
				bool found = false;
				for (int i = 0; i < banks.Count; i++)
					if (banksizes[i] + item.Value.Size <= bankmax)
					{
						banks[i].Add(item);
						banksizes[i] += item.Value.Size;
						found = true;
						break;
					}
				if (!found)
				{
					banks.Add(new List<KeyValuePair<string, MusicInfo>>() { item });
					banksizes.Add(item.Value.Size);
				}
			}
			using (StreamWriter sw = new StreamWriter("musicbanks.asm", false, Encoding.ASCII))
			{
				sw.WriteLine("MusicPointerList macro");
				foreach (var item in songs)
				{
					sw.Write("\tzmakePlaylistEntry\tMus_{0},", item.Key);
					switch (Path.GetExtension(item.Value.OutputFile ?? item.Value.File).ToLowerInvariant())
					{
						case ".asm":
						case ".68k":
							if (item.Value.PALMode)
								sw.WriteLine("musprop_uncompressed|musprop_palmode");
							else
								sw.WriteLine("musprop_uncompressed");
							break;
						case ".bin":
							if (item.Value.PALMode)
								sw.WriteLine("musprop_palmode");
							else
								sw.WriteLine("0");
							break;
					}
				}
				sw.WriteLine("\tendm");
				sw.WriteLine();
				for (int i = 0; i < banks.Count; i++)
				{
					sw.WriteLine("; ------------------------------------------------------------------------------");
					sw.WriteLine("; Music bank {0}", i + 1);
					sw.WriteLine("; ------------------------------------------------------------------------------");
					sw.WriteLine("SndMus{0}_Start:	startBank", i + 1);
					if (i == 0)
						sw.WriteLine("zMasterPlaylist:");
					sw.WriteLine("\tMusicPointerList");
					sw.WriteLine();
					foreach (var item in banks[i])
						switch (Path.GetExtension(item.Value.OutputFile ?? item.Value.File).ToLowerInvariant())
						{
							case ".asm":
							case ".68k":
								sw.WriteLine("Mus_{0}:\tinclude \"sound/music/{1}\" ; ${2:X} bytes", item.Key, item.Value.File, item.Value.Size);
								break;
							case ".bin":
								sw.WriteLine("Mus_{0}:\tBINCLUDE \"sound/music/{1}\" ; ${2:X} bytes", item.Key, item.Value.OutputFile ?? item.Value.File, item.Value.Size);
								break;
						}
					sw.WriteLine();
					sw.WriteLine("\tfinishBank");
					sw.WriteLine();
				}
			}
			using (StreamWriter sw = new StreamWriter("musicbanklist.asm", false, Encoding.ASCII))
			{
				sw.WriteLine("zMusicBanks:");
				foreach (var item in songs)
					sw.WriteLine("\tmakeMusicBank {0}", item.Key);
			}
			using (StreamWriter sw = new StreamWriter("musicids.asm", false, Encoding.ASCII))
			{
				List<string> musids = new List<string>(songs.Keys);
				musids.Add("_End");
				sw.WriteLine("MusID__First = 1");
				string last = "_First";
				for (int i = 0; i < musids.Count; i++)
				{
					if (i % 7 == 0)
						sw.Write("\tenum MusID_{0}=MusID_{1}", musids[i], last);
					else
						sw.Write(",MusID_{0}", musids[i]);
					if (i % 7 == 6)
					{
						sw.WriteLine();
						last = musids[i] + "+1";
					}
				}
			}
			using (StreamWriter sw = new StreamWriter("musicnames.asm", false, Encoding.ASCII))
			{
				sw.WriteLine("SongNames:\toffsetTable");
				sw.WriteLine("\toffsetTableEntry.w\tMusNam_Null");
				foreach (var item in songs)
					sw.WriteLine("\toffsetTableEntry.w\tMusNam_{0}", item.Key);
				sw.WriteLine();
				sw.WriteLine("MusNam_Null:\tdc.b 0,' '");
				foreach (var item in songs)
					sw.WriteLine("MusNam_{0}:\tsongtext\t\"{1}\"", item.Key, item.Value.Title.ToUpperInvariant());
				sw.WriteLine("\teven");
			}
			var playlistini = IniFile.IniFile.Load("musiclist.ini");
			// IniSerializer can't process a Dictionary with values as single-line collections, so we have to do it manually.
			var playlist = new Dictionary<string, string[]>(playlistini.Count);
			foreach (var item in playlistini[string.Empty])
				playlist[item.Key] = item.Value.Split(new[] { ", " }, StringSplitOptions.None);
			using (StreamWriter sw = new StreamWriter("musiclist.asm", false, Encoding.ASCII))
			{
				sw.WriteLine("MusicList:\toffsetTable");
				foreach (var item in playlist)
					if (item.Key != "SBZ")
						sw.WriteLine("MusLstPtr_{0}:\toffsetTableEntry.w\tMusicList_{0}", item.Key);
				sw.WriteLine();
				int numsbz = playlist["SBZ"].Length;
				foreach (var item in playlist)
				{
					if (item.Key == "SBZ") continue;
					sw.WriteLine("MusicList_{0}:\tdc.b MusicList_{0}_End-MusicList_{0}_Start", item.Key);
					sw.WriteLine("MusicList_{0}_Start:", item.Key);
					for (int i = 0; i < item.Value.Length; i++)
						item.Value[i] = "MusID_" + item.Value[i];
					sw.WriteLine("\tdc.b {0}", string.Join(", ", item.Value));
					if (item.Key == "LZ")
					{
						sw.WriteLine("MusicList_SBZ_Start:");
						for (int i = 0; i < playlist["SBZ"].Length; i++)
							playlist["SBZ"][i] = "MusID_" + playlist["SBZ"][i];
						sw.WriteLine("\tdc.b {0}", string.Join(", ", playlist["SBZ"]));
					}
					sw.WriteLine("MusicList_{0}_End:", item.Key);
					sw.WriteLine();
					if (item.Key != "LZ")
						Console.WriteLine("{0} has {1} songs.", item.Key, item.Value.Length);
					else
						Console.WriteLine("{0} has {1} songs, {2} for LZ and {3} for SBZ.", item.Key, item.Value.Length + numsbz, item.Value.Length, numsbz);
				}
				sw.WriteLine("\teven");
			}
		}
	}

	class MusicInfo
	{
		public string Title { get; set; }
		public string Author { get; set; }
		public string File { get; set; }
		public bool PALMode { get; set; }
		[DefaultValue(true)]
		public bool Compressible { get; set; }
		public bool OfsInName { get; set; }	// for KEHmusic-smpsplay
		public short Size { get; set; }
		public string MD5 { get; set; }
		public string OutputFile { get; set; }
	}

	class DACInfo
	{
		public string File { get; set; }
		[TypeConverter(typeof(ByteHexConverter))]
		public byte Delay { get; set; }
	}

	public class ByteHexConverter : TypeConverter
	{
		public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
		{
			if (destinationType == typeof(string))
				return true;
			return base.CanConvertTo(context, destinationType);
		}

		public override object ConvertTo(ITypeDescriptorContext context, CultureInfo culture, object value, Type destinationType)
		{
			if (destinationType == typeof(string) && value is byte)
				return ((byte)value).ToString("X");
			return base.ConvertTo(context, culture, value, destinationType);
		}

		public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
		{
			if (sourceType == typeof(string))
				return true;
			return base.CanConvertFrom(context, sourceType);
		}

		public override object ConvertFrom(ITypeDescriptorContext context, CultureInfo culture, object value)
		{
			if (value is string)
				return byte.Parse((string)value, NumberStyles.HexNumber);
			return base.ConvertFrom(context, culture, value);
		}

		public override bool IsValid(ITypeDescriptorContext context, object value)
		{
			if (value is byte)
				return true;
			byte i;
			if (value is string)
				return byte.TryParse((string)value, NumberStyles.HexNumber, NumberFormatInfo.InvariantInfo, out i);
			return base.IsValid(context, value);
		}
	}
}
